from ._ensure_model import ensure_model, ensure_user_model


def active_membership(user, team):
    """
    Gets the user's membership to the given team,
    if they have one currently.

    :param user:    The user.
    :param team:    The team to get the membership to.
    :return:        The membership.
    """
    # Local imports to avoid circular dependencies
    from ..models import Team, Membership

    # Make sure the user argument is a user
    ensure_user_model(user, False)

    # Make sure the team argument is a team
    ensure_model(team, Team)

    # Get the active memberships between this user and the team
    # (should be zero or one)
    memberships = Membership.objects.active().filter(user=user, team=team)

    # Check the previous condition (very bad if it fails)
    num_memberships: int = len(memberships)
    if num_memberships > 1:
        raise AssertionError(f"Each user should have at most one active membership with "
                             f"a team, but {user} has {num_memberships} with {team}")

    # If no active memberships, return None
    if num_memberships == 0:
        return None

    # Otherwise there is one and only one, so return it
    return memberships[0]



from django.conf import settings
from django.db import models

from ..apps import APP_NAME
from ..mixins import TeamOwnedModel, SoftDeleteModel, SoftDeleteQuerySet
from ..util import ensure_user_model


class TeamQuerySet(SoftDeleteQuerySet):
    """
    Custom query-set for teams, which can filter results to
    those that a given user has admin access to.
    """
    def user_is_admin_for(self, user):
        """
        Filters the teams to those the given user is an
        admin for.

        :param user:    The user.
        :return:        The teams.
        """
        # Local import to avoid circular references
        from ._Membership import Membership

        # Make sure we are passed a user model
        ensure_user_model(user)

        # Unauthenticated users can't be admins
        if not user.is_authenticated:
            return self.none()

        # Superusers/staff are honourary admins for all teams
        if user.is_superuser or user.is_staff:
            return self.all()

        return self.filter(memberships__permissions=Membership.PERMISSION_ADMIN,
                           memberships__deletion_time__isnull=True,
                           memberships__user=user)


class Team(TeamOwnedModel, SoftDeleteModel):
    """
    A team represents a collection of users.
    """
    # The name of the team
    name = models.CharField(max_length=200)

    # The members of the team
    members = models.ManyToManyField(settings.AUTH_USER_MODEL,
                                     through=f"{APP_NAME}.Membership",
                                     through_fields=("team", "user"),
                                     related_name="teams")

    class Meta(SoftDeleteModel.Meta):
        constraints = [
            # Ensure that each active organisation has a unique name
            models.UniqueConstraint(name="unique_active_organisation_names",
                                    fields=["name"],
                                    condition=SoftDeleteModel.active_Q)
        ]

    def get_owning_team(self):
        return self

    def pre_delete(self):
        # Delete the memberships to this team
        self.memberships.all().delete()

    def pre_delete_bulk(self, query_set):
        # Delete the memberships to this team
        from ._Membership import Membership
        Membership.objects.filter(team__in=query_set).delete()

    def __str__(self) -> str:
        return self.name



# Generated by Django 2.2.6 on 2019-12-03 03:50

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Membership',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('creation_time', models.DateTimeField(auto_now_add=True)),
                ('deletion_time', models.DateTimeField(default=None, editable=False, null=True)),
                ('permissions', models.CharField(choices=[('R', 'Read'), ('W', 'Write'), ('A', 'Admin')], default='R', max_length=1)),
                ('creator', models.ForeignKey(editable=False, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
                'base_manager_name': 'objects',
                'default_manager_name': 'objects',
            },
        ),
        migrations.CreateModel(
            name='Team',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('creation_time', models.DateTimeField(auto_now_add=True)),
                ('deletion_time', models.DateTimeField(default=None, editable=False, null=True)),
                ('name', models.CharField(max_length=200)),
                ('creator', models.ForeignKey(editable=False, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('members', models.ManyToManyField(related_name='teams', through='simple_django_teams.Membership', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
                'base_manager_name': 'objects',
                'default_manager_name': 'objects',
            },
        ),
        migrations.AddField(
            model_name='membership',
            name='team',
            field=models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='memberships', to='simple_django_teams.Team'),
        ),
        migrations.AddField(
            model_name='membership',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='memberships', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddConstraint(
            model_name='team',
            constraint=models.UniqueConstraint(condition=models.Q(deletion_time__isnull=True), fields=('name',), name='unique_active_organisation_names'),
        ),
        migrations.AddConstraint(
            model_name='membership',
            constraint=models.UniqueConstraint(condition=models.Q(deletion_time__isnull=True), fields=('user', 'team'), name='one_active_membership_per_user_per_team'),
        ),
    ]



from typing import Type

from django.db.models import Model
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser


def ensure_model(value, model: Type[Model]):
    """
    Checks that the given value is an instance of a particular model.

    :param value:   The value to check.
    :param model:   The model type to check for.
    """
    if not isinstance(value, model):
        raise TypeError(f"Expected {model.__name__} but got {value.__class__.__name__} instead")


def ensure_user_model(value, allow_anonymous: bool = True):
    """
    Checks that the given value is a user.

    :param value:               The value to check.
    :param allow_anonymous:     Whether to also allow anonymous users.
    """
    if not allow_anonymous or not isinstance(value, AnonymousUser):
        ensure_model(value, get_user_model())



from django.conf import settings
from django.db import models
from django.utils.timezone import now


class SoftDeleteQuerySet(models.QuerySet):
    """
    Base class for query-sets over soft-delete models.
    """
    def active(self):
        """
        Filters the query-set to only those that are active.
        """
        return self.filter(SoftDeleteModel.active_Q)

    def deleted(self):
        """
        Filters the query-set to those that are deleted.
        """
        return self.filter(SoftDeleteModel.deleted_Q)

    def delete(self):
        # Only delete active models
        active = self.active()

        # Perform any pre-delete actions
        active.model.pre_delete_bulk(active)

        # Set the deletion time of the active models
        active.update(deletion_time=now())


class SoftDeleteModel(models.Model):
    """
    Mixin for models that shouldn't be deleted from the database,
    instead recording a deletion time.
    """
    # The user that created the object
    creator = models.ForeignKey(settings.AUTH_USER_MODEL,
                                on_delete=models.DO_NOTHING,
                                related_name="+",
                                editable=False)

    # The creation date/time of the object
    creation_time = models.DateTimeField(auto_now_add=True,
                                         editable=False)

    # The deletion time of the object. A value of null means it hasn't been deleted
    deletion_time = models.DateTimeField(null=True,
                                         default=None,
                                         editable=False)

    # Q object for filtering active and deleted models
    active_Q = models.Q(deletion_time__isnull=True)
    deleted_Q = models.Q(deletion_time__isnull=False)

    # Manager for soft-delete models
    objects = SoftDeleteQuerySet.as_manager()

    class Meta:
        abstract = True
        base_manager_name = default_manager_name = "objects"

    def is_active(self) -> bool:
        """
        Whether the model is still active (not deleted).
        """
        return self.deletion_time is None

    def is_deleted(self) -> bool:
        """
        Whether the model has been deleted.\
        """
        return not self.is_active()

    def pre_delete(self):
        """
        Can be overridden to perform other actions before an instance
        of the model is deleted.
        """
        pass

    def pre_delete_bulk(self, query_set):
        """
        Can be overridden to perform other actions before a set
        of models are deleted.

        :param query_set:   The set of objects being deleted.
        """
        pass

    def delete(self, using=None, keep_parents=False):
        # Can't delete the deleted
        if self.is_deleted():
            return

        # Perform pre-delete actions
        self.pre_delete()

        # Set the deletion time
        self.deletion_time = now()

        self.save(update_fields=["deletion_time"])
